#!/usr/bin/env python
# -*- coding: utf-8 -*-

import ast
import os
import re
from itertools import izip_longest

import polib

from c3po.conf import settings
from c3po.converters.unicode import UnicodeWriter, UnicodeReader


METADATA_EMPTY = "{'comment': '', 'previous_msgctxt': None, " + \
                 "'encoding': 'utf-8', 'obsolete': 0, 'msgid_plural': '', " + \
                 "'msgstr_plural': {}, 'occurrences': [], 'msgctxt': None, " + \
                 "'flags': [], 'previous_msgid': None, " + \
                 "'previous_msgid_plural': None}"


def _get_all_po_filenames(locale_root, lang, po_files_path):
    """
    Get all po filenames from locale folder and return list of them.
    Assumes a directory structure:
    <locale_root>/<lang>/<po_files_path>/<filename>.
    """
    all_files = os.listdir(os.path.join(locale_root, lang, po_files_path))
    return filter(lambda s: s.endswith('.po'), all_files)


def _get_new_csv_writers(trans_title, meta_title,
                         trans_csv_path, meta_csv_path):
    """
    Prepare new csv writers, write title rows and return them.
    """
    trans_writer = UnicodeWriter(trans_csv_path)
    trans_writer.writerow(trans_title)

    meta_writer = UnicodeWriter(meta_csv_path)
    meta_writer.writerow(meta_title)

    return trans_writer, meta_writer


def _prepare_locale_dirs(languages, locale_root):
    """
    Prepare locale dirs for writing po files.
    Create new directories if they doesn't exist.
    """
    trans_languages = []
    for i, t in enumerate(languages):
        lang = t.split(':')[0]
        trans_languages.append(lang)
        lang_path = os.path.join(locale_root, lang)
        if not os.path.exists(lang_path):
            os.makedirs(lang_path)
    return trans_languages


def _prepare_polib_files(files_dict, filename, languages,
                         locale_root, po_files_path, header):
    """
    Prepare polib file object for writing/reading from them.
    Create directories and write header if needed. For each language,
    ensure there's a translation file named "filename" in the correct place.
    Assumes (and creates) a directory structure:
    <locale_root>/<lang>/<po_files_path>/<filename>.
    """
    files_dict[filename] = {}
    for lang in languages:
        file_path = os.path.join(locale_root, lang, po_files_path)
        if not os.path.exists(file_path):
            os.makedirs(file_path)

        if header is not None:
            _write_header(os.path.join(file_path, filename), lang, header)

        files_dict[filename][lang] = polib.pofile(
            os.path.join(file_path, filename), encoding="UTF-8")


def _write_entries(po_files, languages, msgid, msgstrs, metadata, comment):
    """
    Write msgstr for every language with all needed metadata and comment.
    Metadata are parser from string into dict, so read them only from gdocs.
    """
    start = re.compile(r'^[\s]+')
    end = re.compile(r'[\s]+$')
    for i, lang in enumerate(languages):
        meta = ast.literal_eval(metadata)
        entry = polib.POEntry(**meta)
        entry.tcomment = comment
        entry.msgid = msgid
        if msgstrs[i]:
            start_ws = start.search(msgid)
            end_ws = end.search(msgid)
            entry.msgstr = str(start_ws.group() if start_ws else '') + \
                unicode(msgstrs[i].strip()) + \
                str(end_ws.group() if end_ws else '')
        else:
            entry.msgstr = ''
        po_files[lang].append(entry)


def _write_header(po_path, lang, header):
    """
    Write header into po file for specific lang.
    Metadata are read from settings file.
    """
    po_file = open(po_path, 'w')
    po_file.write(header + '\n')
    po_file.write(
        'msgid ""' +
        '\nmsgstr ""' +
        '\n"MIME-Version: ' + settings.METADATA['MIME-Version'] + r'\n"'
        '\n"Content-Type: ' + settings.METADATA['Content-Type'] + r'\n"'
        '\n"Content-Transfer-Encoding: ' +
        settings.METADATA['Content-Transfer-Encoding'] + r'\n"'
        '\n"Language: ' + lang + r'\n"' + '\n')
    po_file.close()


def _write_new_messages(po_file_path, trans_writer, meta_writer,
                        msgids, msgstrs, languages):
    """
    Write new msgids which appeared in po files with empty msgstrs values
    and metadata. Look for all new msgids which are diffed with msgids list
    provided as an argument.
    """
    po_filename = os.path.basename(po_file_path)
    po_file = polib.pofile(po_file_path)

    new_trans = 0
    for entry in po_file:
        if entry.msgid not in msgids:
            new_trans += 1
            trans = [po_filename, entry.tcomment, entry.msgid, entry.msgstr]
            for lang in languages[1:]:
                trans.append(msgstrs[lang].get(entry.msgid, ''))

            meta = dict(entry.__dict__)
            meta.pop('msgid', None)
            meta.pop('msgstr', None)
            meta.pop('tcomment', None)

            trans_writer.writerow(trans)
            meta_writer.writerow([str(meta)])

    return new_trans


def _get_new_msgstrs(po_file_path, msgids):
    """
    Write new msgids which appeared in po files with empty msgstrs values
    and metadata. Look for all new msgids which are diffed with msgids list
    provided as an argument.
    """
    po_file = polib.pofile(po_file_path)

    msgstrs = {}

    for entry in po_file:
        if entry.msgid not in msgids:
            msgstrs[entry.msgid] = entry.msgstr

    return msgstrs


def po_to_csv_merge(languages, locale_root, po_files_path,
                    local_trans_csv, local_meta_csv,
                    gdocs_trans_csv, gdocs_meta_csv):
    """
    Converts po file to csv GDocs spreadsheet readable format.
    Merges them if some msgid aren't in the spreadsheet.
    :param languages: list of language codes
    :param locale_root: path to locale root folder containing directories
                        with languages
    :param po_files_path: path from lang directory to po file
    :param local_trans_csv: path where local csv with translations
                            will be created
    :param local_meta_csv: path where local csv with metadata will be created
    :param gdocs_trans_csv: path to gdoc csv with translations
    """
    msgids = []

    trans_reader = UnicodeReader(gdocs_trans_csv)
    meta_reader = UnicodeReader(gdocs_meta_csv)

    try:
        trans_title = trans_reader.next()
        meta_title = meta_reader.next()
    except StopIteration:
        trans_title = ['file', 'comment', 'msgid']
        trans_title += map(lambda s: s + ':msgstr', languages)
        meta_title = ['metadata']

    trans_writer, meta_writer = _get_new_csv_writers(
        trans_title, meta_title, local_trans_csv, local_meta_csv)

    for trans_row, meta_row in izip_longest(trans_reader, meta_reader):
        msgids.append(trans_row[2])
        trans_writer.writerow(trans_row)
        meta_writer.writerow(meta_row if meta_row else [METADATA_EMPTY])

    trans_reader.close()
    meta_reader.close()

    po_files = _get_all_po_filenames(locale_root, languages[0], po_files_path)

    new_trans = False
    for po_filename in po_files:
        new_msgstrs = {}
        for lang in languages[1:]:
            po_file_path = os.path.join(locale_root, lang,
                                        po_files_path, po_filename)
            if not os.path.exists(po_file_path):
                open(po_file_path, 'a').close()
            new_msgstrs[lang] = _get_new_msgstrs(po_file_path, msgids)

        if len(new_msgstrs[languages[1]].keys()) > 0:
            new_trans = True
            po_file_path = os.path.join(locale_root, languages[0],
                                        po_files_path, po_filename)
            _write_new_messages(po_file_path, trans_writer, meta_writer,
                                msgids, new_msgstrs, languages)

    trans_writer.close()
    meta_writer.close()

    return new_trans


def csv_to_po(trans_csv_path, meta_csv_path, locale_root,
              po_files_path, header=None):
    """
    Converts GDocs spreadsheet generated csv file into po file.
    :param trans_csv_path: path to temporary file with translations
    :param meta_csv_path: path to temporary file with meta information
    :param locale_root: path to locale root folder containing directories
                        with languages
    :param po_files_path: path from lang directory to po file
    """
    pattern = "^\w+.*po$"
    for root, dirs, files in os.walk(locale_root):
        for f in filter(lambda x: re.match(pattern, x), files):
            os.remove(os.path.join(root, f))

    # read title row and prepare descriptors for po files in each lang
    trans_reader = UnicodeReader(trans_csv_path)
    meta_reader = UnicodeReader(meta_csv_path)
    try:
        title_row = trans_reader.next()
    except StopIteration:
        # empty file
        return

    trans_languages = _prepare_locale_dirs(title_row[3:], locale_root)

    po_files = {}

    meta_reader.next()
    # go through every row in downloaded csv file
    for trans_row, meta_row in izip_longest(trans_reader, meta_reader):
        filename = trans_row[0].rstrip()
        metadata = meta_row[0].rstrip() if meta_row else METADATA_EMPTY
        comment = trans_row[1]
        msgid = trans_row[2]

        if filename not in po_files:
            _prepare_polib_files(po_files, filename, trans_languages,
                                 locale_root, po_files_path, header)

        _write_entries(po_files[filename], trans_languages, msgid,
                       trans_row[3:], metadata, comment)
    for filename in po_files:
        for lang in po_files[filename]:
            po_files[filename][lang].save()

    trans_reader.close()
    meta_reader.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*

import os
import subprocess
import urlparse
from subprocess import Popen, PIPE

import gdata.spreadsheet.service
import gdata.docs.client
import gdata.docs.data
import gdata.data
import gdata.docs.service
from gdata.client import RequestError

from c3po.conf import settings
from c3po.converters.po_csv import csv_to_po, po_to_csv_merge
from c3po.converters.po_ods import po_to_ods, csv_to_ods


LOCAL_ODS = 'local.ods'
GDOCS_TRANS_CSV = 'c3po_gdocs_trans.csv'
GDOCS_META_CSV = 'c3po_gdocs_meta.csv'
LOCAL_TRANS_CSV = 'c3po_local_trans.csv'
LOCAL_META_CSV = 'c3po_local_meta.csv'


class PODocsError(Exception):
    pass


class Communicator(object):
    """
    Client for communicating with GDocs. Providing log in on object
    creation and methods for synchronizing, uploading,
    downloading files and clearing GDoc.

    Needs to specify:
        locale_root, po_files_path
    For example in Django, where 'en' is language code:
        conf/locale/en/LC_MESSAGES/django.po
        conf/locale/en/LC_MESSAGES/custom.po
    locale_root='conf/locale/'
    po_files_path='LC_MESSAGES'
    """

    email = None
    password = None
    url = None
    source = None
    temp_path = None
    languages = None
    locale_root = None
    po_files_path = None
    header = None

    def __init__(self, email=None, password=None, url=None, source=None,
                 temp_path=None, languages=None, locale_root=None,
                 po_files_path=None, header=None):
        """
        Initialize object with all necessary client information and log in
        :param email: user gmail account address
        :param password: password to gmail account
        :param url: url to spreadsheet where translations are or will be placed
        :param temp_path: path where temporary files will be saved
        :param source: source information to show on web
        :param languages: list of languages
        :param locale_root: path to locale root folder containing directories
                            with languages
        :param po_files_path: path from lang directory to po file
        :param header: header which will be put on top of every po file when
                       downloading
        """
        construct_vars = ('email', 'password', 'url', 'source', 'temp_path',
                          'languages', 'locale_root', 'po_files_path', 'header')
        for cv in construct_vars:
            if locals().get(cv) is None:
                setattr(self, cv, getattr(settings, cv.upper()))
            else:
                setattr(self, cv, locals().get(cv))
        self._login()
        self._get_gdocs_key()
        self._ensure_temp_path_exists()

    def _login(self):
        """
        Login into Google Docs with user authentication info.
        """
        try:
            self.gd_client = gdata.docs.client.DocsClient()
            self.gd_client.ClientLogin(self.email, self.password, self.source)
        except RequestError as e:
            raise PODocsError(e)

    def _get_gdocs_key(self):
        """
        Parse GDocs key from Spreadsheet url.
        """
        try:
            args = urlparse.parse_qs(urlparse.urlparse(self.url).query)
            self.key = args['key'][0]
        except KeyError as e:
            raise PODocsError(e)

    def _ensure_temp_path_exists(self):
        """
        Make sure temp directory exists and create one if it does not.
        """
        try:
            if not os.path.exists(self.temp_path):
                os.mkdir(self.temp_path)
        except OSError as e:
            raise PODocsError(e)

    def _clear_temp(self):
        """
        Clear temp directory from created csv and ods files during
        communicator operations.
        """
        temp_files = [LOCAL_ODS, GDOCS_TRANS_CSV, GDOCS_META_CSV,
                      LOCAL_TRANS_CSV, LOCAL_META_CSV]
        for temp_file in temp_files:
            file_path = os.path.join(self.temp_path, temp_file)
            if os.path.exists(file_path):
                os.remove(file_path)

    def _download_csv_from_gdocs(self, trans_csv_path, meta_csv_path):
        """
        Download csv from GDoc.
        :return: returns resource if worksheets are present
        :except: raises PODocsError with info if communication
                 with GDocs lead to any errors
        """
        try:
            entry = self.gd_client.GetResourceById(self.key)
            self.gd_client.DownloadResource(
                entry, trans_csv_path,
                extra_params={'gid': 0, 'exportFormat': 'csv'}
            )
            self.gd_client.DownloadResource(
                entry, meta_csv_path,
                extra_params={'gid': 1, 'exportFormat': 'csv'}
            )
        except (RequestError, IOError) as e:
            raise PODocsError(e)
        return entry

    def _upload_file_to_gdoc(
            self, file_path,
            content_type='application/x-vnd.oasis.opendocument.spreadsheet'):
        """
        Uploads file to GDocs spreadsheet.
        Content type can be provided as argument, default is ods.
        """
        try:
            entry = self.gd_client.GetResourceById(self.key)
            media = gdata.data.MediaSource(
                file_path=file_path, content_type=content_type)
            self.gd_client.UpdateResource(
                entry, media=media, update_metadata=True)
        except (RequestError, IOError) as e:
            raise PODocsError(e)

    def _merge_local_and_gdoc(self, entry, local_trans_csv,
                              local_meta_csv, gdocs_trans_csv, gdocs_meta_csv):
        """
        Download csv from GDoc.
        :return: returns resource if worksheets are present
        :except: raises PODocsError with info if communication with GDocs
                 lead to any errors
        """
        try:
            new_translations = po_to_csv_merge(
                self.languages, self.locale_root, self.po_files_path,
                local_trans_csv, local_meta_csv,
                gdocs_trans_csv, gdocs_meta_csv)
            if new_translations:
                local_ods = os.path.join(self.temp_path, LOCAL_ODS)
                csv_to_ods(local_trans_csv, local_meta_csv, local_ods)
                media = gdata.data.MediaSource(
                    file_path=local_ods, content_type=
                    'application/x-vnd.oasis.opendocument.spreadsheet'
                )
                self.gd_client.UpdateResource(entry, media=media,
                                              update_metadata=True)
        except (IOError, OSError, RequestError) as e:
            raise PODocsError(e)

    def synchronize(self):
        """
        Synchronize local po files with translations on GDocs Spreadsheet.
        Downloads two csv files, merges them and converts into po files
        structure. If new msgids appeared in po files, this method creates
        new ods with appended content and sends it to GDocs.
        """
        gdocs_trans_csv = os.path.join(self.temp_path, GDOCS_TRANS_CSV)
        gdocs_meta_csv = os.path.join(self.temp_path, GDOCS_META_CSV)
        local_trans_csv = os.path.join(self.temp_path, LOCAL_TRANS_CSV)
        local_meta_csv = os.path.join(self.temp_path, LOCAL_META_CSV)

        try:
            entry = self._download_csv_from_gdocs(gdocs_trans_csv,
                                                  gdocs_meta_csv)
        except PODocsError as e:
            if 'Sheet 1 not found' in str(e) \
                    or 'Conversion failed unexpectedly' in str(e):
                self.upload()
            else:
                raise PODocsError(e)
        else:
            self._merge_local_and_gdoc(entry, local_trans_csv, local_meta_csv,
                                       gdocs_trans_csv, gdocs_meta_csv)

            try:
                csv_to_po(local_trans_csv, local_meta_csv,
                          self.locale_root, self.po_files_path, self.header)
            except IOError as e:
                raise PODocsError(e)

        self._clear_temp()

    def download(self):
        """
        Download csv files from GDocs and convert them into po files structure.
        """
        trans_csv_path = os.path.realpath(
            os.path.join(self.temp_path, GDOCS_TRANS_CSV))
        meta_csv_path = os.path.realpath(
            os.path.join(self.temp_path, GDOCS_META_CSV))

        self._download_csv_from_gdocs(trans_csv_path, meta_csv_path)

        try:
            csv_to_po(trans_csv_path, meta_csv_path,
                      self.locale_root, self.po_files_path, header=self.header)
        except IOError as e:
            raise PODocsError(e)

        self._clear_temp()

    def upload(self):
        """
        Upload all po files to GDocs ignoring conflicts.
        This method looks for all msgids in po_files and sends them
        as ods to GDocs Spreadsheet.
        """
        local_ods_path = os.path.join(self.temp_path, LOCAL_ODS)
        try:
            po_to_ods(self.languages, self.locale_root,
                      self.po_files_path, local_ods_path)
        except (IOError, OSError) as e:
            raise PODocsError(e)

        self._upload_file_to_gdoc(local_ods_path)

        self._clear_temp()

    def clear(self):
        """
        Clear GDoc Spreadsheet by sending empty csv file.
        """
        empty_file_path = os.path.join(self.temp_path, 'empty.csv')
        try:
            empty_file = open(empty_file_path, 'w')
            empty_file.write(',')
            empty_file.close()
        except IOError as e:
            raise PODocsError(e)

        self._upload_file_to_gdoc(empty_file_path, content_type='text/csv')

        os.remove(empty_file_path)


def git_push(git_message=None, git_repository=None, git_branch=None,
             locale_root=None):
    """
    Pushes specified directory to git remote
    :param git_message: commit message
    :param git_repository: repository address
    :param git_branch: git branch
    :param locale_root: path to locale root folder containing directories
                        with languages
    :return: tuple stdout, stderr of completed command
    """
    if git_message is None:
        git_message = settings.GIT_MESSAGE
    if git_repository is None:
        git_repository = settings.GIT_REPOSITORY
    if git_branch is None:
        git_branch = settings.GIT_BRANCH
    if locale_root is None:
        locale_root = settings.LOCALE_ROOT

    try:
        subprocess.check_call(['git', 'checkout', git_branch])
    except subprocess.CalledProcessError:
        try:
            subprocess.check_call(['git', 'checkout', '-b', git_branch])
        except subprocess.CalledProcessError as e:
            raise PODocsError(e)

    try:
        subprocess.check_call(['git', 'ls-remote', git_repository])
    except subprocess.CalledProcessError:
        try:
            subprocess.check_call(['git', 'remote', 'add', 'po_translator',
                                   git_repository])
        except subprocess.CalledProcessError as e:
            raise PODocsError(e)

    commands = 'git add ' + locale_root + \
               ' && git commit -m "' + git_message + '"' + \
               ' && git push po_translator ' + git_branch + ':' + git_branch
    proc = Popen(commands, shell=True, stdout=PIPE, stderr=PIPE)
    stdout, stderr = proc.communicate()

    return stdout, stderr


def git_checkout(git_branch=None, locale_root=None):
    """
    Checkouts branch to last commit
    :param git_branch: branch to checkout
    :param locale_root: locale folder path
    :return: tuple stdout, stderr of completed command
    """
    if git_branch is None:
        git_branch = settings.GIT_BRANCH
    if locale_root is None:
        locale_root = settings.LOCALE_ROOT

    proc = Popen('git checkout ' + git_branch + ' -- ' + locale_root,
                 shell=True, stdout=PIPE, stderr=PIPE)
    stdout, stderr = proc.communicate()

    return stdout, stderr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import cStringIO
import csv


class UTF8Recoder(object):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, fd, encoding):
        self.reader = codecs.getreader(encoding)(fd)

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode("utf-8")


class UnicodeReader(object):
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, file_path, dialect=csv.excel,
                 encoding="utf-8", **kwargs):
        self.file = open(file_path, 'rb')
        recorder = UTF8Recoder(self.file, encoding)
        self.reader = csv.reader(recorder, dialect=dialect, **kwargs)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self

    def close(self):
        self.file.close()


class UnicodeWriter(object):
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, file_path, dialect=csv.excel,
                 encoding="utf-8", **kwargs):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwargs)
        self.stream = open(file_path, 'wb')
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

    def close(self):
        self.stream.close()
